import click
import os
import yaml
import pickle
import io
import numpy as np
import math
import subprocess
import time
import re
import logging
from collections import defaultdict
#import .FPGA_Execution.base_functions as base_functions
from cri_simulations.config import *
import cri_simulations.FPGA_Execution.base_functions as base_functions
from cri_simulations.FPGA_Execution.fpga_compiler import text2script

#n_internal = None
#n_inputs = None
#inputs = None
ng_num = 16
debug_prints = False

def bin_format(integer, length):
    """converts an integer to a binary string of length length
    """
    return f'{integer:0>{length}b}'


def load(fromFile = False, n_internal=None, n_inputs=None, inputs=None):
    """Reads configuration values from yaml file.

    Populates the global variables n_internal, n_inputs, and inputs with
    the values of the corresponding keys in config.yaml. n_internal is the number
    of neurons in the network, n_inputs is the number of axons, inputs is a list of
    inputs to the network at different time steps. We may no longer use the input values
    read from the yaml file.

    """
    #global n_internal
    #global n_inputs
    #global inputs

    if fromFile:
        if os.path.isfile('FPGA_Execution/config.yaml'):
            with open("FPGA_Execution/config.yaml", 'r') as f:
                data = yaml.safe_load(f)

            n_internal = data['n_internal']
            n_inputs = data['n_inputs']
            inputs = data['inputs']
        else:
            raise FileNotFoundError('File Not Found, or an Error was detected. Have you ran the --init command yet?')
    else:
            n_internal = n_internal
            n_inputs = n_inputs
            inputs = inputs   

def twos_comp(val, bits):
    """compute the 2's complement of int value val

    Takes the int casting of a two's compliment binary string and the number
    of bytes in the binary string and returns the value of the two's compliment
    representation correspondin to the original binary string.

    Parameters
    ----------
    val : int
        the binary stirng to compute the two's compliment of casted to an int
    bits : int
        the number of bits corresponding to the original binary representation of val

    Return
    ------
    val : int
        the value of the original two's compliment binary string
         
    """
    if (val & (1 << (bits - 1))) != 0: # if sign bit is set e.g., 8bit: 128-255
        val = val - (1 << bits)        # compute negative value
    return val                         # return positive value as is

def get_packet_type(string):
    """Caluclates and prints the membrane potentials corresponding to a PCIe packet from
       the FPGA

    Given the packet returned by a PCIe read from URAM this function calculates the neuron index
    and membrane potential of the neruon represented by the PCIe packet and prints those values
    to terminal

    Parameters
    ----------
    string : str
        The string returned by the PCIe command to read URAM

    """
    if debug_prints:
        print(string)
    #breakpoint()
    string = string.split('\n')[3]
    string = string.replace('INFO', '')
    new = ''
    for i in string:
    
        if i not in [':', '?', '.', '@']:
            new += i

    new = new.strip().split(' ')
    new = [i for i in list(filter(None, new)) if len(i) < 3]
    #breakpoint()
    tag = new[15]+new[14]
    return tag
    #print(new)

def read_spikes(string):
    """Caluclates and prints the incoming spikes corresponding to a PCIe packet from
       the FPGA

    Given a spike packet this function decodes the spike packet

    Parameters
    ----------
    string : str
        The string returned by the PCIe command to read spikes

    """
    if debug_prints:
        print(string)
    splitString = string.split('\n')
    processedString = []
    #breakpoint()
    for substring in splitString:
        if (substring): #ignore the empty stirng split generates after the last newline
            strippedSubstring = substring.replace('INFO', '')
            new = ''
            for i in strippedSubstring:
                #breakpoint()
                if i in ['0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F', ' ']:
                    new += i
            new = new.strip().split(' ')
            new = [i for i in list(filter(None, new)) if len(i) < 3]
            processedString += new
    processedString.reverse()
    mergedString = "".join(processedString)
    spikeInt = int(mergedString, 16) #convert hex to dec
    binLength = len(mergedString)*4 #calculate number of bits hex string should occupy
    binary = bin_format(spikeInt, binLength) #convert to binary string
    #extract the tag
    tag = '{:0{width}x}'.format(int(binary[:32], 2), width=2)
    #extract the executionRun_counter
    executionRun_counter = int(binary[-32:], 2)
    #extract the region containing spike data
    spikeData = binary[32:-32]
    spikePacketLength = 32
    spikeList = []
    for spikePacket in [spikeData[i:i+spikePacketLength] for i in range(0, len(spikeData), spikePacketLength)]:
        subexecutionRun_counter, address = processSpikePacket(spikePacket)
        if (subexecutionRun_counter != None and address != None):
            spikeList.append((subexecutionRun_counter,address))
    #breakpoint()
    #print('spikeList: ' + str(spikeList))
    return executionRun_counter, spikeList

def processSpikePacket(spikePacket):
    """Processes an incoming binary string representing a single spike event
    """
    #breakpoint()
    valid = bool(int(spikePacket[8])) #check if it's a valid spike packet
    #valid = True #TODO: THIS IS JUST FOR DEBUGGING DELETE THIS LINE
    if valid:
        subexecutionRun_counter = int(spikePacket[0:8], 2)
        address = int(spikePacket[-17:],2)
        #breakpoint()
        return subexecutionRun_counter, address
    else:
        return None, None

def read_membranes(string):
    """Caluclates and prints the membrane potentials corresponding to a PCIe packet from
       the FPGA

    Given the packet returned by a PCIe read from URAM this function calculates the neuron index
    and membrane potential of the neruon represented by the PCIe packet and prints those values
    to terminal

    Parameters
    ----------
    string : str
        The string returned by the PCIe command to read URAM

    """
    #breakpoint()
    #if debug_prints:
    logging.debug(string)
    substring = string.partition('\n')[0]
    strippedSubstring = substring.replace('INFO', '')
    new = ''
    for i in strippedSubstring:
    
        if i not in [':', '?', '.', '@']:
            new += i

    new = new.strip().split(' ')
    new = [i for i in list(filter(None, new)) if len(i) < 3]


    #print(new)
    col = bin(int(new[6],16))[2:].zfill(8)
    #print(col)
    col = int(col[3:7],2)
    #print(col)
    row =  (bin(int(new[6],16))[2:].zfill(8))[7] + bin(int(new[5],16))[2:].zfill(8) + (bin(int(new[4],16))[2:].zfill(8))[:4]
    row = int(row,2)
    #print(row)
    mp = (bin(int(new[4],16))[2:].zfill(8))[4:] + bin(int(new[3],16))[2:].zfill(8) + bin(int(new[2],16))[2:].zfill(8) + bin(int(new[1],16))[2:].zfill(8) + bin(int(new[0],16))[2:].zfill(8)
    mp = twos_comp(int(mp,2), len(mp))
    #print(mp)
    #input("wait")
    #print('Neuron index %d, Neuron Row %d, Neuron Column %d, Membrane Potential %d'% (ng_num * row + col, row, col, mp))
    return (ng_num * row + col, row, col, mp)
    #input("press enter")
'''    while new:
        #Reverse the order
        mp = new[1] + new[0]
        addr = new[4] + new[3] + new[2]
        mp = bin(int(mp,16))[2:].zfill(16)
        addr = bin(int(addr,16))[2:].zfill(17)
        addr = addr[-17:] #only use last 17 bits
        mp = twos_comp(int(mp,2), len(mp))
#        print('row,col', int(addr[3:],2), int(addr[0:3],2))
        print('Neuron index %d, Membrane Potential %d'% (8 * int(addr[3:],2) + int(addr[0:3],2), mp))
        del new[0:5]
'''

@click.group()
def interface():
    """Untested
    """
    pass

def flush_reads():
    """Flushes any remaining reads from the FPGA

    Notes
    -----
    This function currently does not work and appears to cause issues with successive operations interacting
    with the FPGA

    """
    command = 'sudo adxdma_dmadump rb 0 0 0x40'
    while True:
        try:
            subprocess.run(command,shell=True, timeout=10, check=True)
        except:
            print("timeout expired")
            break

def clear_address_packet(row,col,simDump=False):
    rowBin = np.binary_repr(row,13)
    colBin = np.binary_repr(col,4)
    byte4 = rowBin[-4:]+'0'*4
    byte5 = rowBin[-12:-4]
    byte6 = '0'*2+'1'+np.binary_repr(col,4)+rowBin[0]

    byte4_hex = '{:0{width}x}'.format(int(byte4, 2), width=2)
    byte5_hex = '{:0{width}x}'.format(int(byte5, 2), width=2)
    byte6_hex = '{:0{width}x}'.format(int(byte6, 2), width=2)

    if simDump:
        packet = '00 00 00 00' + ' ' + byte4_hex + ' ' +  byte5_hex + ' ' + byte6_hex 
    else:
        packet = '0x00 0x00 0x00 0x00' + ' 0x' + byte4_hex + ' 0x' +  byte5_hex + ' 0x' + byte6_hex 
    return packet

#@interface.command()
def clear(n_internal, simDump = False, coreID = 0):
    """This function clears the membrane potentials on the fpga.

    """
    #updated Feb11
    #load()#is this even needed
    coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
    coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)
    commandTail = ' 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 ' + coreByte + ' 03'
    if simDump:
        dump = []
    for i in range(int(np.ceil(n_internal / ng_num))):
        if simDump:
            numCol = 16
            for column in range(numCol):
                #command = "03" + "00"*56 + byte + "00" + '{:0{width}x}'.format(i, width=2) + "00"*4
                command = clear_address_packet(row=i,col=column,simDump=True) + commandTail
                command = command.split(sep=' ')
                command.reverse()
                command = ''.join(command)
                dump.append(command)

            #0 0 0 0 $(($i * 16)) 0x00 0x20 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0x0 0x0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 3

        else:
            command = 'sudo adxdma_dmadump wb 0 0x0 '
            numCol = 16
            for column in range(numCol):
                if (column == numCol-1):
                    command = command + clear_address_packet(row=i,col=column) + commandTail #we don't need the trailing space for the last line of the command
                else:
                    command = command + clear_address_packet(row=i,col=column) + commandTail + ' '

            os.system(command)
            #print(command)
            time.sleep(1)

    if simDump:
        return dump

#@interface.command()

def read_address_packet(row,col, simDump=False):
    rowBin = np.binary_repr(row,13)
    colBin = np.binary_repr(col,4)
    byte4 = rowBin[-4:]+'0'*4
    byte5 = rowBin[-12:-4]
    byte6 = '0'*3+np.binary_repr(col,4)+rowBin[0]

    byte4_hex = '{:0{width}x}'.format(int(byte4, 2), width=2)
    byte5_hex = '{:0{width}x}'.format(int(byte5, 2), width=2)
    byte6_hex = '{:0{width}x}'.format(int(byte6, 2), width=2)

    if simDump:
        packet = '00 00 00 00' + ' ' + byte4_hex + ' ' +  byte5_hex + ' ' + byte6_hex
    else:
        packet = '0x00 0x00 0x00 0x00' + ' 0x' + byte4_hex + ' 0x' +  byte5_hex + ' 0x' + byte6_hex 
    return packet

def clear_read_buffer(coreID = 0):
    while(True):
        currentRead = subprocess.run(['sudo', 'adxdma_dmadump', 'rb', '0', '0' ,'0x40'], stdout=subprocess.PIPE, check=True).stdout.decode('utf-8')
        if(get_packet_type(currentRead) == 'FFFF'):
            break

def flush_spikes(coreID = 0):
    spikeOutput = []
    while(True):
        currentRead = subprocess.run(['sudo', 'adxdma_dmadump', 'rb', '0', '0' ,'0x40'], stdout=subprocess.PIPE, check=True).stdout.decode('utf-8')
        if(get_packet_type(currentRead) == 'FFFF'):
            break
        elif(get_packet_type(currentRead) == 'EEEE'): #TODO: this is just here for debugging
            executionRun_counter, spikeList = read_spikes(currentRead)
            spikeOutput = spikeOutput + spikeList
        else:
            logging.error("None spike packet encountered during spike flush")

    return spikeOutput



def read(n_internal, simDump = False, coreID = 0):
    """This function reads in the membrane potentials from the fpga.

    """
    #breakpoint()
    above_thresh = []
    #updated Feb11
    #load()#is this even needen
    coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
    coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)
    commandTail = ' 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 ' + coreByte + ' 03'
    if simDump:
        dump = []
    formated_results = [] #holds the formated output
    #spikeOutput = []
    for i in range(int(np.ceil(n_internal/ng_num))):
        if simDump:
            numCol = 16
            for column in range(numCol):
                #command = "03" + "00"*56 + byte + "00" + '{:0{width}x}'.format(i, width=2) + "00"*4
                command = read_address_packet(row=i,col=column,simDump=True) + commandTail
                command = command.split(sep=' ')
                command.reverse()
                command = ''.join(command)
                dump.append(command)
            #for i in range(ng_num):
            #    dump.append("40")
        else:
            command = 'sudo adxdma_dmadump wb 0 0x0 '
            numCol = 16
            for column in range(numCol):
                #we'll split the command into lines to make it a little easier to read if we choose to print it
                if (column == numCol-1):
                    command = command + read_address_packet(row=i,col=column) + commandTail #we don't need the slash or new line for the last line of the command

                else:
                    command = command + read_address_packet(row=i,col=column) + commandTail + ' '
            subprocess.run(command, shell=True, check=True)

            #time.sleep(1)
            result = [] #holds the raw pcie output

            for i in range(ng_num):
                #print("NEURON GROUP NUMBER: "+str(i))
                while True: #look for the next membrane potential in the fifo
                    currentRead = subprocess.run(['sudo', 'adxdma_dmadump', 'rb', '0', '0' ,'0x40'], stdout=subprocess.PIPE, check=True).stdout.decode('utf-8')
                    if(get_packet_type(currentRead) == 'CCCC'):
                        break
                    else: #TODO: this is just here for debugging
                        #executionRun_counter, spikeList = read_spikes(currentRead)
                        #spikeOutput = spikeOutput + spikeList
                        logging.error("an errant packet was detected during membrane potential readout")
                        #print(executionRun_counter)
                        #print(spikeList)

                result.append(currentRead)                
                #breakpoint()
                #time.sleep(1)
                #print(result[i])
                formated_results.append(read_membranes(currentRead))
            #print(command)
            
    if simDump:
        return dump
    else:
        return formated_results#, spikeOutput


@interface.command()
@click.option('--filename',default=None, help='Pickled Network Configuration file')
@click.option('--n_inputs', default=None, help='Number of axons')
@click.option('--n_internal',default=None, help='Number of Neurons')
def init(filename,n_inputs,n_internal):
    """Untested
    """
    if filename is None or n_inputs is None or n_internal is None:
        raise ValueError('Command Line Arguments must be specified')


    with open(filename, 'rb' ) as f:
        file = pickle.load(f)

    inp = file[0]

    try:
        n_inputs = int(n_inputs)
        n_internal = int(n_internal)
    except ValueError:
        print('Failed! Make sure the number of inputs and internal values are numbers')

    data = {'inputs' : inp, 'n_inputs': n_inputs, 'n_internal': n_internal}
    with io.open('config.yaml', 'w', encoding='utf8') as outfile:
        yaml.dump(data, outfile, default_flow_style=False, allow_unicode=True)

##PCIe commands for setting network parameters
## [511:504] --> 0x4
## [71:70] --> exec_neuron_model
## [69:34] --> 36-bit threshold
## [33:17] --> NUM OUTPUTs in the network
## [16:0] --> NUM INPUTs in the network

def write_parameters(neron_model, threshold, n_outputs, n_inputs, coreID = 0, simDump = False):
    """Writes the network parameters to the FPGA

    Parameters
    ----------
    neuron_model : int
        The type of neuron model to use (1: incremental I&F, 2: leaky I&F, 3: non-leaky I&F)
    threshold : int
        Neuron spike threshold
    n_outputs : int
        The number of neurons in the network
    n_inputs : int
        The number of axons in the network
    """
    #load()
    n_inputs_bin = np.binary_repr(n_inputs,17)
    a_bin = n_inputs_bin[9:17]
    b_bin = n_inputs_bin[1:9]
    c_bin = ["0"] * 8
    c_bin[7] = n_inputs_bin[0] #set lsb of c_bin
    n_outputs_bin = np.binary_repr(n_outputs,17)
    c_bin[:7] = n_outputs_bin[10:17] #set msb of cbit as lsb of outputs
    c_bin = "".join(c_bin)
    d_bin = n_outputs_bin[2:10]
    e_bin = ["0"] * 8
    e_bin[6:8] = n_outputs_bin[:2]
    threshold_bin = np.binary_repr(threshold,36)
    e_bin[:6] = threshold_bin[30:]
    e_bin = "".join(e_bin)
    f_bin = threshold_bin[22:30]
    g_bin = threshold_bin[14:22]
    h_bin = threshold_bin[6:14]
    i_bin = ["0"] * 8
    i_bin[2:] = threshold_bin[:6]
    n_model_bin = np.binary_repr(neron_model,2)
    i_bin[:2] = n_model_bin
    i_bin = "".join(i_bin)
    h_hex = "0x04"
    a_hex = '{:0{width}x}'.format(int(a_bin, 2), width=2)
    b_hex = '{:0{width}x}'.format(int(b_bin, 2), width=2)
    c_hex = '{:0{width}x}'.format(int(c_bin, 2), width=2)
    d_hex = '{:0{width}x}'.format(int(d_bin, 2), width=2)
    e_hex = '{:0{width}x}'.format(int(e_bin, 2), width=2)
    f_hex = '{:0{width}x}'.format(int(f_bin, 2), width=2)
    g_hex = '{:0{width}x}'.format(int(g_bin, 2), width=2)
    h_hex = '{:0{width}x}'.format(int(h_bin, 2), width=2)
    i_hex = '{:0{width}x}'.format(int(i_bin, 2), width=2)
    #print(a_hex)
    #print(c_hex)
    #print(threshold_bin)

    coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
    coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)
    if simDump:
        return ["04"+coreByte+"00"*53+i_hex+h_hex+g_hex+f_hex+e_hex+d_hex+c_hex+b_hex+a_hex]

    command = "sudo adxdma_dmadump wb 0 0x0 0x$a 0x$b 0x$c 0x$d 0x$e 0x$f 0x$g 0x$h 0x$i 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 \
                  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 " +"0x"+coreByte +" 0x04"
    command = command.replace('$a', a_hex)
    command = command.replace('$b', b_hex)
    command = command.replace('$c', c_hex)
    command = command.replace('$d', d_hex)
    command = command.replace('$e', e_hex)
    command = command.replace('$f', f_hex)
    command = command.replace('$g', g_hex)
    command = command.replace('$h', h_hex)
    command = command.replace('$i', i_hex)
    logging.info("programminng network parameters")
    #print(command)
    os.system(command)

#@interface.command()
def execute(simDump=False, coreID = 0):
    """ Runs a single step of the network and prints "executing a Timestep: " to the terminal
    """
    # Implement Running the Command for one timestep
    coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
    coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)
    if simDump:
        command = ["06"+coreByte+"00"*62]
        return command
    else:
        #click.echo('Executing a Timestep: ')
        base_functions.execute(coreID)

@interface.command()
@click.option('--filename', default=' ', help='HBM configuration file')
def configure(filename):
    """Untested
    """
    # Implement Configuring the network
    click.echo('Starting Board Configuration')
    if not os.path.isfile(filename):
        raise FileNotFoundError('File "%s" does not exist'%(filename))

    os.system('sudo chmod +x %s'%(filename))
    os.system('sudo ./%s'%(filename))

@interface.command()
@click.option('--timestep',default=0,help='Timestep Number')
def input(timestep, n_inputs, inputs):
    """Untesetd
    """

    # Implement Configuring the network
    click.echo('Sending inputs at timestep %d to the board'%(timestep))
    #load()
    #print(inputs)
    vals = inputs[timestep]

    for i in vals:
        if i >= n_inputs or i < 0:
            raise ValueError('An input is not valid (Not in the number of external inputs)')

    binary = '0' * n_inputs
    binary = "".join(['1' if i in vals else '0' for i in range(n_inputs)])[::-1]
    final = hex(int(binary, 2))

    #Send write external input PCIe command
    os.system('sudo adxdma_dmadump wb 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1')
    os.system('sudo adxdma_dmadump wb 0 0 %s 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0'%(final))

def step_input(command):
    """This just runs a command formated to be run with bash
    """
    os.system(command)

def input_user(inputs, numAxons, simDump = False, coreID=0, reserve=False):
    """Generates the input command for a given time step

    Generates the necesary bash command to run to provide inputs to the network for a given timestep

    Parameters
    ----------
    inputs : list of int
        The currently spiking axons

    Returns
    -------
    command : str
        The bash command to run to send the input to the FPGA
    """
    if simDump:
        if (reserve):
            dump = []
            command = "01"+"00"*63
            currInput = inputs
                #opCode = "0x01"

            one_hot_bin = ["0"] * 256
            for axon in currInput:
                one_hot_bin[axon] = "1"
            coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
            coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)
            one_hot_bin = one_hot_bin

            one_hot_bin.reverse() #makesure LSB is on left
            one_hot_bin_reverse = "".join(one_hot_bin) #make sure LSB is on the left
            one_hot = "00"+coreByte+31*"00"+'{:0{width}x}'.format(int(one_hot_bin_reverse, 2), width=31*2)
            dump.append(command)
            dump.append(one_hot)
        else:
            dump = []
            command = "01"+"00"*63
            dump.append(command)
            currInput = inputs
            currInput.sort()
            for count in range(math.ceil(numAxons/512)):
                #opCode = "0x01"
                one_hot_bin = ["0"] * 512
                inputSegment = [i for i in currInput if (512*count) <= i and i < (512*count+512) ]
                for axon in inputSegment:
                    one_hot_bin[axon%512] = "1"

                one_hot_bin.reverse() #makesure LSB is on left
                one_hot_bin_reverse = "".join(one_hot_bin) #make sure LSB is on the left
                one_hot = '{:0{width}x}'.format(int(one_hot_bin_reverse, 2), width=64*2)
            #dump.append(command)
                dump.append(one_hot)
        #breakpoint()
        return dump

    else:
        if (reserve): #new format upper 256 bits are reserved
            #hex_list = []
            command = "sudo adxdma_dmadump wb 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1"
            #print(self.input)
            currInput = inputs
            #opCode = "0x01"
            one_hot_bin = ["0"] * 256
            for axon in currInput:
                one_hot_bin[axon] = "1"
            #one_hot_bin = one_hot_bin[::-1]
            while one_hot_bin:
                curr_byte = one_hot_bin[:8][::-1]
                curr_byte = "".join(curr_byte)
                command = command + " 0x"+'{:0{width}x}'.format(int(curr_byte, 2), width=2)
                one_hot_bin = one_hot_bin[8:]
            tail = 30*" 0x00"
            coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
            coreByte = "0x"+'{:0{width}x}'.format(int(coreBits, 2), width=2)
            command = command+tail+" "+coreByte +" "+"0x00"
        else:
            #hex_list = []
            
            #print(self.input)
            currInput = inputs
            currInput.sort()
            command = "sudo adxdma_dmadump wb 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1"
            #opCode = "0x01"
            for count in range(math.ceil(numAxons/512)):
                
                one_hot_bin = ["0"] * 512
                inputSegment = [i for i in currInput if (512*count) <= i and i < (512*count+512) ]
                for axon in inputSegment:
                    one_hot_bin[axon%512] = "1"
            #one_hot_bin = one_hot_bin[::-1]
                while one_hot_bin:
                    curr_byte = one_hot_bin[:8][::-1]
                    curr_byte = "".join(curr_byte)
                    command = command + " 0x"+'{:0{width}x}'.format(int(curr_byte, 2), width=2)
                    one_hot_bin = one_hot_bin[8:]
        #breakpoint()
        subprocess.run(command,shell = True, check = True)
    
@interface.command()
def fullrun():
   """Untested
   """
   os.system('sudo ./hyddenn2_new.sh 0')
   base_functions.execute()
   base_functions.flush()

def execute_step():
   """Duplicate of execute() without printing to terminal
   """
   base_functions.execute()

@interface.command()
@click.option('--nums',default=1,help='The number of the commands to flush')
def flush(nums):
    """Untested
    """
    base_functions.flush(nums)

def write_synapse_row(r, row, simDump = False, coreID = 0):
    '''
    Creates the necessary adxdma_dump commands to program the synapses into HBM

    Returns
    -------
    script : str
        The bash commands to run to program the synapses in HBM
    '''
    #breakpoint()
    commandPrefix = '02' + '{:0{width}x}'.format(coreID, width=2) + '000000000000000000000000000000000000000000000000000000'
    if simDump:
        dump = []
    #weights = self.synapses
    #script = ''
    #n = 0
    #for r, d in enumerate(weights):
    cmd = ''
    for w in row:
        if w[0] == 0:
            # [31] = 0 for internal connections and 1 for external connections, [30:29] = unused for single core
            # TODO: how do I know if a given synapse is an internal or external connection?
            cmd += '{:0{width}x}'.format(int(np.binary_repr( 0, SYN_OP_BITS) + np.binary_repr(int(w[1]), SYN_ADDR_BITS) + np.binary_repr(int(w[2]),SYN_WEIGHT_BITS), 2),width=8)
        #TODO: looks like this format is out of date, it should be the same as the above format. It may not be hyper critical
        elif w[0] == 1:
            spike = str(w[0]) + 15 * '0'
            addr = np.binary_repr(w[1], SYN_ADDR_BITS)
            #cmd += '{:0{width}x}'.format(int(spike + addr, 2), width=8)
            cmd += '{:0{width}x}'.format(int(np.binary_repr( 4, SYN_OP_BITS) + 12*'0' + np.binary_repr(w[1],17), 2),width=8)



    # append HBM write opcode and address
    #breakpoint()
    cmd = commandPrefix + '{:0{width}x}'.format(0x800000 + SYN_BASE_ADDR + r, width=6) + cmd
    # append command to complete script list
    if simDump:
        dump.append(cmd)
    else:
        hexString = text2script(cmd)
        finalCmd = HBM_WRITE_CMD + hexString[:-1]
        #
    #n = n + 1

    # write to text file
    if simDump:
        return dump
    else:
        os.system(finalCmd)


if __name__ == '__main__':
    interface()



import os
import subprocess
from collections import defaultdict
import re
import numpy as np

def execute(coreID):
    """This function runs a single time step of the network

    """
    coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
    coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)

    subprocess.run('sudo adxdma_dmadump wb 0 0x0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0x0 0x0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 '+coreByte+' 6', shell = True, check = True)


def flush(nums):
    """Untested
    """
    result = []
    for i in range(nums):
        result.append(
            (subprocess.run(['sudo', 'adxdma_dmadump', 'rb', '0', '0', '64'], stdout=subprocess.PIPE)).stdout.decode(
                'utf-8'))
    # print(result,type(result))

    output = re.findall("INFO:.*:([\s0-9A-F]*)", result[0])
    output = [i.strip() for i in output]
    spikes = defaultdict(list)
    for i in range(len(output)):
        data = output[i].split()
        # Get the timestep information
        while data:
            if (data[0] == 'EF' and data[1] == 'CD' and data[2] == 'AB' and data[3] == '89') or (
                    data[0] == 'EE' and data[1] == 'EE'):
                pass
            else:
                # working solution for now
                timestep = int(data[3], 16)
                spike = data[2] + data[1] + data[0]
                spike = int(spike, 16)
                if spike not in spikes[timestep]:
                    spikes[timestep].append(spike)
            del data[0:4]

    #print(spikes)



import numpy as np
import pickle
import os
import sys
import warnings
import logging
import numpy as np
#warnings.filterwarnings('error')

def text2script(cmd_str):
    """Converts a string of hex characters into the correct format to suppyl to the adxdma dump command.

    Given a string of hex characters with the left most character containing the MSB create a string of pairs of hex characters representing bytes
    with the leftmost byte contanining the LSB in the format expected by the adxdma_dmadump binary for the data argument.

    Parameters
    ----------
    cmd_str : str
        The string of hexidecimal characters to format. The first character represents the hex character containing the MSB

    Returns
    -------
    script_txt : str
        The formated string of bytes
    """
    cmd = cmd_str
    script_txt = ''
    while cmd:
        script_txt += '0x' + cmd[-2:] + ' '
        cmd = cmd[0:-2]
    return script_txt


class fpga_compiler:
    """Produces the needed adxdma dump scripts of a given network to program HBM

    Attributes
    ----------
    input : dict
        The inputs to the network for each timestep. Key, timestep value, list of axons
    axon_ptrs : numpy array
        Array of tuples pointing to the rows containing the synapses for the corresponding Axon.
    Each tuple is (start row, end row).
    neuron_ptrs : numpy array
        Array of tuples pointing to the rows containing the synapses for the corresponding Axon.
    Each tuple is (start row, end row).
    synapses : list
        List of tuples corresponding to synapses. Each tuple is (oncore/offcore bit, synapse address (row index of destination neuron pointer in HBM calculated as floor(destination neuron index / number of neuron groups)), weight)
    HBM_WRITE_CMD : str

    HBM_OP_RW  : str
        OP code to read/write to hbm vie PCie
    NRN_BASE_ADDR : int
        Starting address of neuron pointers in HBM
    SYN_BASE_ADDR : int
        Starting address of synapses in HBM
    PTR_ADDR_BITS : int
        Number of bits used to represent pointer starting adderess
    PTR_LEN_BITS :
        Number of bits used to represent the number of rows of synapses a pointer corresponds to
    SYN_OP_BITS :
        Number of bits used to represent synapse opcode
    SYN_ADDR_BITS :
        Number of bits used to represent synapse address
    SYN_WEIGHT_BITS :
        Number of bits used to represent synapse weight

    """

    # OP code to read/write to hbm vie PCie
    #HBM_OP_RW = '02' + 28 * '00'

    PTR_ADDR_BITS = 23  # HBM row address
    PTR_LEN_BITS = 9  # (8 connections/row x 2^9 rows = 2^12 = 4K connections max)
    AXN_BASE_ADDR = 0  # axon pointer start address
    NRN_BASE_ADDR = 2 ** 14  # neuron pointer start address (2^17 axons / 8 axon pointers/row = 2^14 rows max)
    SYN_BASE_ADDR = 2 ** 15  # (2^17 neurons / 8 neuron pointers/row = 2^14 rows max -> 2^14+2^14=2^15)

    SYN_ADDR_BITS = 13  # each synapse: [31:29]=OpCode  [28:16]=address, [15:0]=weight (1 sign + 15 value, fixed-point)
    SYN_WEIGHT_BITS = 16
    SYN_OP_BITS = 3

    HBM_WRITE_CMD = 'sudo adxdma_dmadump wb 0 0 '

    WRITE = '01' + 63 * '00'

    def __init__(self, data, N_neurons, outputs, coreID = 0):
        '''
        Creates the FPGA Compiler object. Populates spike packets in synapses.

        Paramaters
        ----------
        data: list
            Takes the following format: [0]=input data, [1]=axon pointers, [2]=neuron pointers, [3]=synapses
        N_neurons : int
            Number of neurons in network
        CoreID : int
            Index of core in FPGA to write to
        '''
        coreBits = np.binary_repr(coreID,5)+3*'0'#5 bits for coreID
        coreByte = '{:0{width}x}'.format(int(coreBits, 2), width=2)
        self.HBM_OP_RW = '02' + coreByte + 27 * '00'
        #self.input = data[0]
        self.axon_ptrs = data[0]
        self.neuron_ptrs = data[1]
        self.synapses = data[2]

        n = 0
        errors = 0
        #This works because of how python handles references but it would be much clearer
        #to directly work with self.neuron_ptrs and self.synapses
        #Find space to insert spike packets and insert them in the numpy array representing synapses

        #only assign spike entries in HBM for neurons in the outputs list
        #for row in range(len(data[1])):
            #for col in range(len(data[1][row]))
        rowLength = len(data[1][0])
        for neuronIdx in outputs:
            row = int(np.floor(neuronIdx / rowLength))
            col = neuronIdx % rowLength
            start = data[1][row][col][0]
            end = data[1][row][col][1] + 1
            finished = False
            for r in data[2][start:end]:
                if r[3] == (0,0,0) and n <= len(outputs):
                    r[3] = (1, neuronIdx)
                    finished = True
                    break
                for i in range(len(r)):
                    if r[i] == (0, 0, 0) and n <= len(outputs):
                        r[i] = (1, neuronIdx)
                        #n += 1
                        finished = True
                        break

                if finished:
                    break
            if not finished:
                if n == len(outputs):
                    pass
                else:
                    
                    if n > N_neurons:
                        logging.info('neuron pointer index is greater than specified number of internal neurons, a spike entry will not be assigned to remaining neuron pointer')
                    else:
                         errors+=1
                         logging.error('neuron pointer index: '+ str(n) +' could not be assigned a spike entry')



        logging.info('spike entry assignment errors: ', errors)
        #print(self.synapses)
        #sys.exit()



    def create_axon_ptrs(self, simDump = False):
        '''
        Creates the necessary adxdma_dump commands to program the axon pointers into HBM

        Returns
        -------
        script : str
            The bash commands to run to program the axon pointers in HBM
        '''

        if simDump:
            dump = []
        axn_ptrs = np.fliplr(self.axon_ptrs)
        #print('axn_ptrs')
        #print(axn_ptrs)

        script = ''
        for r, d in enumerate(axn_ptrs):

            cmd = ''
            for p in d:
                decimal_addr = int(
                    np.binary_repr(p[1] - p[0], self.PTR_LEN_BITS) + np.binary_repr(p[0] + self.SYN_BASE_ADDR,
                                                                                    self.PTR_ADDR_BITS), 2)

                cmd += '{:0{width}x}'.format(decimal_addr, width=8)


            # append HBM write opcode and WRITE command
            cmd = self.HBM_OP_RW + '{:0{width}x}'.format(0x800000 + r, width=6) + cmd
            if simDump:
                #print("we made itttttttttt")
                dump.append(cmd)
            # append command to complete script list
            script += self.txt2script(cmd) + '\n'
        if simDump:
            return dump
        else:
            return script

    def txt2script(self, cmd_str):
        """Converts a string of hex characters into the correct format to suppyl to the adxdma dump command.

        Given a string of hex characters with the left most character containing the MSB create a string of pairs of hex characters representing bytes
        with the leftmost byte contanining the LSB in the format expected by the adxdma_dmadump binary for the data argument.

        Parameters
        ----------
        cmd_str : str
            The string of hexidecimal characters to format. The first character represents the hex character containing the MSB

        Returns
        -------
        script_txt : str
            The formated string of bytes
        """
        cmd = cmd_str
        script_txt = ''
        while cmd:
            script_txt += '0x' + cmd[-2:] + ' '
            cmd = cmd[0:-2]
        return script_txt

    def create_neuron_ptrs(self, simDump = False):
        '''
        Creates the necessary data arguments to pass to the adxdma_dump commands to program the neuron pointers into HBM.
        Data arguments for multiple adxdma_dump commands are seperated by new line characters

        Returns
        -------
        script : str
            The data arguments to provide to a series of adxdma_dump commands. Data arguments for successive adxdma_dump commands
        are seperated by newline characters
        '''
        if simDump:
            dump = []
        #print(self.neuron_ptrs)
        nrn_ptrs = np.fliplr(self.neuron_ptrs)
        script = ''
        for r, d in enumerate(nrn_ptrs):
            cmd = ''
            for p in d:
                #print(d)
                #print(p)
                cmd += '{:0{width}x}'.format(
                    int(np.binary_repr(p[1] - p[0], self.PTR_LEN_BITS) + np.binary_repr(p[0] + self.SYN_BASE_ADDR,
                                                                                        self.PTR_ADDR_BITS), 2),
                    width=8)


            # append HBM write opcode and address
            cmd = self.HBM_OP_RW + '{:0{width}x}'.format(0x800000 + self.NRN_BASE_ADDR + r, width=6) + cmd

            # append command to complete script list
            if simDump:
                dump.append(cmd)
            else:
                script += self.txt2script(cmd) + '\n'

        if simDump:
            return dump
        else:
            return script

    def create_synapses(self, simDump = False):
        '''
        Creates the necessary adxdma_dump commands to program the synapses into HBM

        Returns
        -------
        script : str
            The bash commands to run to program the synapses in HBM
        '''

        #print(self.synapses, '\n')
        #weights = np.fliplr(self.synapses)
        #print(weights)
        if simDump:
            dump = []
        weights = self.synapses
        script = ''
        n = 0
        for r, d in enumerate(weights):
            cmd = ''
            for w in d:
                if w[0] == 0:
                    # [31] = 0 for internal connections and 1 for external connections, [30:29] = unused for single core
                    # TODO: how do I know if a given synapse is an internal or external connection?
                    cmd += '{:0{width}x}'.format(int(np.binary_repr( 0, self.SYN_OP_BITS) + np.binary_repr(int(w[1]), self.SYN_ADDR_BITS) + np.binary_repr(int(w[2]),self.SYN_WEIGHT_BITS), 2),width=8)
                #TODO: looks like this format is out of date, it should be the same as the above format. It may not be hyper critical
                elif w[0] == 1:
                    #spike = str(w[0]) + 15 * '0'
                    spike = 16 * '0'
                    addr = np.binary_repr(w[1], self.SYN_ADDR_BITS)
                    #cmd += '{:0{width}x}'.format(int(spike + addr, 2), width=8)
                    cmd += '{:0{width}x}'.format(int(np.binary_repr( 4, self.SYN_OP_BITS) + 12*'0' + np.binary_repr(w[1],17), 2),width=8)



            # append HBM write opcode and address
            cmd = self.HBM_OP_RW + '{:0{width}x}'.format(0x800000 + self.SYN_BASE_ADDR + r, width=6) + cmd
            # append command to complete script list
            if simDump:
                dump.append(cmd)
            else:
                script += self.txt2script(cmd) + '\n'

            #
            n = n + 1

        # write to text file
        if simDump:
            return dump
        else:
            return script


    def gen_input(self,time_step):
            #hex_list = []
            command = "sudo adxdma_dmadump wb 0 0 "
    	  #print(self.input)
            currInput = self.input[time_step]
            opCode = "0x01"
            one_hot_bin = ["0"] * 504
            for axon in currInput:
                one_hot_bin[axon] = "1"
            #one_hot_bin = one_hot_bin[::-1]
            while one_hot_bin:
                curr_byte = one_hot_bin[:8][::-1]
                curr_byte = "".join(curr_byte)
                command = command + " 0x"+'{:0{width}x}'.format(int(curr_byte, 2), width=2)
                one_hot_bin = one_hot_bin[8:]
            command = command+" "+opCode
            return command
            
    def gen_input2(self,time_step,simDump = False):
        """Generates the input command for a given time step

        Generates the necesary bash command to run to provide inputs to the network for a given timestep

        Parameters
        ----------
        time_step : int
            The timestep you wish to generate the input command for

        Returns
        -------
        command : str
            The bash command to run to send the input to the FPGA
        """
        if simDump:
            dump = []
            command = "01"+"00"*63
            currInput = self.input[time_step]
                #opCode = "0x01"
            one_hot_bin = ["0"] * 504
            for axon in currInput:
                one_hot_bin[axon] = "1"

            one_hot_bin.reverse() #makesure LSB is on left
            one_hot_bin_reverse = "".join(one_hot_bin) #make sure LSB is on the left
            one_hot = '{:0{width}x}'.format(int(one_hot_bin_reverse, 2), width=64*2)
            dump.append(command)
            dump.append(one_hot)
            return dump

        else:
            #hex_list = []
            command = "sudo adxdma_dmadump wb 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1"
            #print(self.input)
            currInput = self.input[time_step]
            #opCode = "0x01"
            one_hot_bin = ["0"] * 504
            for axon in currInput:
                one_hot_bin[axon] = "1"
            #one_hot_bin = one_hot_bin[::-1]
            while one_hot_bin:
                curr_byte = one_hot_bin[:8][::-1]
                curr_byte = "".join(curr_byte)
                command = command + " 0x"+'{:0{width}x}'.format(int(curr_byte, 2), width=2)
                one_hot_bin = one_hot_bin[8:]
            command = command+" "+"0x00"
            return command

    def create_input_script(self,num_timesteps,n_inputs,filename):


        commands = ['sudo ./hyddenn2_new.sh 31']
        if num_timesteps < len(self.input):
            hex_timesteps = "0x{:02X}".format(num_timesteps)
        commands.append('sudo adxdma_dmadump wb 0 0 %s 0x00 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0x0 0x0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 7'%(hex_timesteps))
        #print(commands)
        #print(self.input)

        for i in range(num_timesteps+1):
            inputs = [0] * (n_inputs + 1)#initialize inputs
            #print(len(inputs))
            #print(self.input[0][-5:])
            for j in range(len(self.input[i])):
                inputs[self.input[i][j]] = 1

            #print(inputs)
            #Create the inputs and group by 8's
            dump = []
            while(inputs):
                tobin = inputs[:8][::-1]
                tobin = ''.join([str(k) for k in tobin])
                tobin = "0x{:02X}".format(int(tobin,2))
                dump.append(tobin)
                del inputs[:8]
                #print(dump)

            n_bits = 64 - len(dump)
            dump = " ".join(dump)
            c = 'sudo adxdma_dmadump wb 0 0 ' + dump + ' 0' * n_bits
            commands.append(c)

        with open(filename + '_input_script.txt', 'w') as f:
            for i in commands:
                f.write(i + '\n')

    def create_script(self, fname, simDump = False):
        """Generates the bash file to program HBM for the current network

        Generates the bash file needed to program the axon pointers, neuron pointers, and synapses into hbm

        Parameters
        ----------
        fname : int
            The filename to write the script to
        """
        #breakpoint()
        axon_ptrs = self.create_axon_ptrs(simDump)
        neuron_ptrs = self.create_neuron_ptrs(simDump)
        synapses = self.create_synapses(simDump)

        if simDump:
            return axon_ptrs, neuron_ptrs, synapses

        #print(axon_ptrs)
        #sys.exit()

        try:
            path = os.path.join(os.getcwd(), fname + '_script.txt')
            os.remove(path)
        except OSError:
            pass

        with open(fname + '_script.txt', 'a') as f2:
            f2.writelines('echo "Configuring Axon Pointers... "\n')

        cmd = self.HBM_WRITE_CMD

        l = 0
        axon_ptr_length = len(axon_ptrs.split('\n'))
        #print('axon_ptr, length', axon_ptr_length)
        for line in axon_ptrs.split('\n'):
            if line:
                with open(fname + '_script.txt', 'a') as f2:
                    f2.writelines(cmd + line[:-1] + '\n')#[:-1] is to get all elements in line except newline character
                    if l == int(axon_ptr_length /4) or l == int(axon_ptr_length /2) or l == int(axon_ptr_length * 3/4):
                        f2.writelines('echo "Finished with %d out of %d"\n' % (l,axon_ptr_length))
            l+=1
        with open(fname + '_script.txt', 'a') as f2:
            f2.writelines('echo "Configuring Neuron Pointers... "\n')

        cmd = self.HBM_WRITE_CMD

        l = 0
        neuron_ptr_length = len(neuron_ptrs.split('\n'))
        for line in neuron_ptrs.split('\n'):
            if line:
                with open(fname + '_script.txt', 'a') as f2:
                    f2.writelines(cmd + line[:-1] + '\n')
                    if l == int(neuron_ptr_length /4) or l == int(neuron_ptr_length /2) or l == int(neuron_ptr_length * 3/4):

                        f2.writelines('echo "Finished with %d out of %d"\n'%(l, neuron_ptr_length))
            l+=1

        with open(fname + '_script.txt', 'a') as f2:
            f2.writelines('echo "Configuring Synapses... "\n')

        cmd = self.HBM_WRITE_CMD

        l = 0
        syn_length = len(synapses.split('\n'))
        for line in synapses.split('\n'):
            if line:
                with open(fname + '_script.txt', 'a') as f2:
                    f2.writelines(cmd + line[:-1] + '\n')
                    if l == int(syn_length /4) or l == int(syn_length /2) or l == int(syn_length * 3/4):

                        f2.writelines('echo "Finished with %d out of %d"\n'%(l, syn_length))
            l+=1

	

def main():
    with open('test' + '.pkl', 'rb') as f:
        data = pickle.load(f)

    f = fpga_compiler(data, 6271)

    f.create_script('test_config')
    #TODO: these values should not be hardcoded
    f.create_input_script(num_timesteps=0,n_inputs=1,filename='test_input')
    #print(f.create_neuron_ptrs().split('\n'))


if __name__ == '__main__':
    main()



import confuse
import os

__all__ = [
    'ARCH',
    'N_NG',
    'PTRS_PER_ROW',
    'DATA_PER_ROW',
    'HBM_OP_RW',
    'PTR_ADDR_BITS',
    'PTR_LEN_BITS',
    'AXN_BASE_ADDR',
    'NRN_BASE_ADDR',
    'SYN_BASE_ADDR',
    'SYN_ADDR_BITS',
    'SYN_WEIGHT_BITS',
    'SYN_OP_BITS',
    'HBM_WRITE_CMD',
    'WRITE',
]

config = confuse.Configuration('cri_simulations', __name__)
config.set_file(os.path.join(os.path.dirname(__file__), 'config.yaml'))

ARCH = config['FPGA_cluster'].get()
N_NG = config['n_ng'].get(int)
PTRS_PER_ROW = config['ptrs_per_row'].get(int)
DATA_PER_ROW = config['data_per_row'].get(int)
HBM_OP_RW = '02' + 28 * '00'

PTR_ADDR_BITS = 23  # HBM row address
PTR_LEN_BITS = 9  # (8 connections/row x 2^9 rows = 2^12 = 4K connections max)
AXN_BASE_ADDR = 0  # axon pointer start address
NRN_BASE_ADDR = 2 ** 14  # neuron pointer start address (2^17 axons / 8 axon pointers/row = 2^14 rows max)
SYN_BASE_ADDR = 2 ** 15  # (2^17 neurons / 8 neuron pointers/row = 2^14 rows max -> 2^14+2^14=2^15)

SYN_ADDR_BITS = 13  # each synapse: [31:29]=OpCode  [28:16]=address, [15:0]=weight (1 sign + 15 value, fixed-point)
SYN_WEIGHT_BITS = 16
SYN_OP_BITS = 3

HBM_WRITE_CMD = 'sudo adxdma_dmadump wb 0 0 '

WRITE = '01' + 63 * '00'



import copy
import logging

class synapse:
    #postsynapticNeuron = None
    #weight = None
    #synapseType = None
    def __init__(self,presynapticNeuron, postsynapticNeuron, weight):
        self.presynapticNeuron = presynapticNeuron
        self.postsynapticNeuron = postsynapticNeuron
        self.weight = weight
        self.synapseType = None
    def get_postsynapticNeuron(self):
        return self.postsynapticNeuron
    def set_synapseType(self):
        #TODO: check for None
        if (self.postsynapticNeuron and self.presynapticNeuron):
            if(self.postsynapticNeuron.get_core() != self.presynapticNeuron.get_core()):
                self.synapseType = "hetero"
            else:
                self.synapseType = "homo"
        else:
            raise Exception("core is None")
    def get_weight(self):
        return self.weight
    def set_weight(self, newWeight):
        self.weight = newWeight
    def __repr__(self):
        return self.obj2string()
    def __str__(self):
        return self.obj2string()
    def obj2string(self):
        return "("+str(self.postsynapticNeuron.get_user_key())+", "+str(self.weight)+", "+str(self.synapseType)+")"

class neuron:
    globalAxonCount = 0 #STATIC total number of axons
    globalNeuronCount = 0 #STATIC total number of nonAxon neurons
    globalCount = 0 #STATIC total number of both axons and neurons

    #globalIdx = None #Index unique among all neurons/axons
    #globalTypeIdx = None #Index unique among neurons of the same type
    #userKey = None
    #coreIdx = None ##Index unique among all neurons/axons in local core
    #coreTypeIdx = None #Index unique among neurons of the same type in local core
    #core = None
    #neuronType = None
    #synapses = None
    def __init__(self, userKey, neuronType, output = False):
        self.userKey = userKey
        self.neuronType = neuronType
        if (neuronType == 'axon'):
            self.output = False
            self.globalTypeIdx = copy.deepcopy(neuron.globalAxonCount)
            neuron.globalAxonCount += 1
        if (neuronType == 'neuron'):
            self.output = output
            self.globalTypeIdx = copy.deepcopy(neuron.globalNeuronCount)
            neuron.globalNeuronCount += 1
        self.globalIdx = copy.deepcopy(neuron.globalCount)
        neuron.globalCount += 1
        self.core = 0
        self.coreTypeIdx = self.globalTypeIdx
        self.coreIdx = self.globalIdx
        self.synapses = []
    @classmethod
    def reset_count(cls):
        #resets counting variables for neuron class
        #breakpoint()
        cls.globalAxonCount = 0
        cls.globalNeuronCount = 0
        cls.globalCount = 0
    def addSynapse(self,postsynapticNeuron,weight):
        newSynapse = synapse(self,postsynapticNeuron,weight)
        self.synapses.append(newSynapse)
    def get_output(self):
        return self.output
    def get_user_key(self):
        return self.userKey
    def get_synapses(self):
        return self.synapses
    def get_neuron_type(self):
        return self.neuronType
    def set_core(self,core):
        self.core = core
    def get_core(self):
        return self.core
    def set_coreIdx(self,coreIdx):
        self.coreIdx = coreIdx
    def get_coreIdx(self):
        return self.coreIdx
    def set_coreTypeIdx(self,coreTypeIdx):
        self.coreTypeIdx = coreTypeIdx
    def get_coreTypeIdx(self):
        return self.coreTypeIdx
    def set_globalIdx(self,globalIdx):
        self.globalIdx = globalIdx
    def get_globalIdx(self):
        return self.globalIdx
    def get_synapse(self, postsynapticKey):
        canidateSynapses = [currSynapse for currSynapse in self.synapses if currSynapse.get_postsynapticNeuron().get_user_key()==postsynapticKey]
        if (len(canidateSynapses) == 0):
            logging.error("Neuron has no synapse to specified postsynaptic neuron")
        else:
            if (len(canidateSynapses) != 1):
                logging.error("Neuron has two synapses to the same postsynaptic neuron")
            else:
                return canidateSynapses[0]
    def set_synapseTypes(self):
        for synapse in self.synapses:
            synapse.set_synapseType()
    def __repr__(self):
        return self.obj2string()
    def __str__(self):
        return self.obj2string()
    def obj2string(self):
        string =  "userID: "+str(self.userKey)+", GlobalID: "+str(self.globalIdx)+", Core Assignment: "+str(self.core)+", Synapses: "
        for currSynapse in self.synapses:
            string = string + str(currSynapse)
        return string

class connectome:
    #connectomeDict = None
    def __init__(self):
        self.connectomeDict = {}
    def addNeuron(self, neuron):
        self.connectomeDict[neuron.userKey] = neuron
    def __repr__(self):
        return self.obj2string()
    def __str__(self):
        return self.obj2string()
    def get_neuron_by_key(self, neuronKey):
        return self.connectomeDict[neuronKey]
    def obj2string(self):
        string = ""
        for key in self.connectomeDict:
            string = string + str(self.connectomeDict[key])+"\n"
        return string
    def get_axons(self):
        axons = {}
        for key in self.connectomeDict:
            if (self.connectomeDict[key].get_neuron_type() == 'axon'):
                axons[key] = self.connectomeDict[key]
        return axons
    def get_neurons(self):
        neurons = {}
        for key in self.connectomeDict:
            if (self.connectomeDict[key].get_neuron_type() == 'neuron'):
                neurons[key] = self.connectomeDict[key]
        return neurons
    def get_merged_neurons(self):
        mergedNeurons = {}
        for key in self.connectomeDict:
            mergedNeurons[self.connectomeDict[key].get_globalIdx()] = self.connectomeDict[key]
        return mergedNeurons

    def get_part_format(self):
        mergedNeurons = self.get_merged_neurons()
        networkConnectivity = []
        for key in mergedNeurons.keys(): #Nishant's partitioning code expects a specific format so we convert to that format
            networkConnectivity.append([copy.deepcopy(synapse.get_postsynapticNeuron().get_globalIdx()) for synapse in mergedNeurons[key].get_synapses()])
        return networkConnectivity
 
    def get_core_outputs_idx(self, core):
         outputs = []
         for key in self.connectomeDict:
            currNeuron = self.connectomeDict[key]
            if (currNeuron.get_output() == True and currNeuron.get_neuron_type() == 'neuron' and currNeuron.get_core() == core) :
                outputs.append(currNeuron.get_coreTypeIdx())
         return outputs
    def get_outputs_idx(self):
         outputs = []
         for key in self.connectomeDict:
            currNeuron = self.connectomeDict[key]
            if (currNeuron.get_output() == True and currNeuron.get_neuron_type() == 'neuron') :
                outputs.append(currNeuron.get_coreTypeIdx())
         return outputs
    def apply_partition(self,membership):
        #breakpoint()
        mergedNeurons = self.get_merged_neurons()
        for key in membership.keys():
            mergedNeurons[key].set_core(membership[key])
        #breakpoint()
        for neuronKey in self.connectomeDict.keys():
            self.connectomeDict[neuronKey].set_synapseTypes()

       

        



import numpy as np
from cri_simulations.config import *
import sys
from ast import literal_eval
from math import ceil
from bidict import bidict
import logging

try:
    from cri_simulations.partitioning_code.hierarchical_partition import partition as netPart
    partLoad = True
except:
    partLoad = False
    logging.warning("partitioning library failed to load, multicore disabled")

################################################################
# Note: ARCH, N_NG, PTRS_PER_ROW, and DATA_PER_ROW are read in #
# from a YAML config. These values are specific to the current #
# itteration of the CRI hardware and should NOT be altered     #
# unless the corresponding parameters in the hardware change   #
################################################################
"""
class mergedNetwork:
    def __init__(self,axons,connections):
        self.axons = axons
        self.connections = connections
        self.mergeDict = {}
        self.symbol2index = {}

    def merge(self):
        #breakpoint()
        #breakpoint()
        axonKeys =  self.axons.keys()
        connectionKeys = self.connections.keys()
        #ensure keys in axon and neuron dicts are mutually exclusive
        #map those keys to indicies
        mapDict = {} #holds maping from symbols to indicies
        #breakpoint();
        mergeDict = {}
        #construct axon dictionary with ordinal numbers as keys
        for idx, symbol in enumerate(axonKeys):
            mergeDict[idx] = self.axons[symbol]
            mapDict[(symbol,'axons')] = idx
        connectionsStartOffset = idx+1
        #construct connections dicitonary with ordinal numbers as keys
        for idx, symbol in enumerate(connectionKeys):
            currIndex = idx+connectionsStartOffset
            mergeDict[currIndex] = self.connections[symbol]
            mapDict[(symbol,'connections')] = currIndex

        self.symbol2index = bidict(mapDict)
        
        #go through and change symbol based postsynaptic neuron values to corresponding index
        for idx in mergeDict:
            for listIdx in range(len(mergeDict[idx])):
                oldTuple = mergeDict[idx][listIdx]
                newTuple = (self.symbol2index[(oldTuple[0],'connections')],oldTuple[1]) #we now the post synaptic neuron is originially from the connections dict because users never define a connection with a postsynaptic axon
                mergeDict[idx][listIdx] = newTuple

        self.mergeDict = mergeDict

        return mergeDict

    def bifurcate_assignments(self,assignments):
        axonAssignments = {}
        neuronAssignments = {}
        
        for index in assignments.keys():
            symbol = self.symbol2index.inverse[index]
            if (symbol[1] == "axons"):
                axonAssignments[symbol[0]] = assignments[index]
            elif (symbol[1] == "connections"):
                neuronAssignments[symbol[0]] = assignments[index]
            else:
                raise Exception("Invalid neuron catagory")
        return axonAssignments,neuronAssignments
"""
sixteen_flag = True
# 4 ways to load into the network

# - simple text file including the source,destination, weight,

# - ONNX format for deep learning models

# - Networkx connectivity graph

# Some format to accomodate the SpheyreNet Models

def generate_local_neurons(axons,connections,axonAssignments , neuornAssignments,n_cores):
    masterDict = {}
    for i in range(n_cores):
        masterDict[i] = {}
        masterDict[i]["axons"] = {}
        masterDict[i]["connections"] = {}
    for axonKey in axons.keys():
        core = axonAssignments[axonKey] #problem, how do we handle axon indicies across cores?
        masterDict[core]["axons"][axonKey] = axons[axonKey]
    for neuronKey in connections.keys():
        core = neuronAssignments[neuronKey] #problem, how do we handle axon indicies across cores?
        masterDict[core]["connections"][neuronKey] = connections[neuronKey]

        


def get_cores():
    """Get's the number of cores to map the network to
    Returns
    -------
    n_cores : int
        The number of cores to map the network to
    """
    # Get the number of cores to map to
    n_cores = 0
    for fpga_cluster_num, fpga_cluster in enumerate(ARCH):
        for fpga_num, fpga in enumerate(fpga_cluster["FPGA"]):
            for core_cluster_num, core_cluster, in enumerate(fpga["Core_cluster"]):
                for core_num in range(int(core_cluster["Cores"])):
                    n_cores += 1
    return n_cores


def load_network(input, connex, output):
    """Loads the network specification.

    This function loads the inputs and connections specified for the network.
    Also determines the number of FPGA cores to be used.

    Parameters
    ----------
    input : str, optional
        Path to file specifying network inputs. (the default is the path in config.yaml)
    connex : str, optional
        Path to file specifying network connections. (the default is the path in config.yaml)

    Returns
    -------
    axons : dict
        Dictionary specifying axons in the network. Key: axon number Value: Synapse Weights
    connections : dict
        Dictionary specifying neurons in the network. Key: Neuron Number Value: Synapse Weights
    inputs : dict
        Dictionary specifying inputs to the network. Key, Time Step Value, axon
    outputs : dict
        TODO: I'm not sure what the outputs are for. I belive it's unused
    ncores : int
        The number of cores peresent in the CRI system

    """
    axons = {}
    connections = {}
    inputs = {}
    outputs = {}

    # Load in the connectivity file
    ax = None
    with open(connex, "r") as f:
        for line in f:
            if not line.startswith("#"):
                if "axons" in line.lower():
                    ax = True
                elif "neurons" in line.lower():
                    ax = False
                else:
                    pre, post = line.split(":")
                    weights = literal_eval(post.strip())
                    weights = [(int(i[0]), float(i[1])) for i in weights]
                    if ax:
                        axons[int(pre.strip())] = weights
                    else:
                        connections[int(pre.strip())] = weights

    # Load in the inputs file
    with open(input, "r") as f:
        for line in f:
            if not line.startswith("#"):
                pre, post = line.split(":")
                inputs[int(pre.strip())] = literal_eval(post.strip())

    with open(output, "r") as f:
        for line in f:
            if not line.startswith("#"):
                pre, post = line.split(":")
                outputs[int(pre.strip())] = literal_eval(post.strip())

    # Get the number of cores to map to
    # n_cores = 0
    # for fpga_cluster_num, fpga_cluster in enumerate(ARCH):
    #    for fpga_num, fpga in enumerate(fpga_cluster['FPGA']):
    #        for core_cluster_num, core_cluster, in enumerate(fpga['Core_cluster']):
    #            for core_num in range(int(core_cluster['Cores'])):
    #                n_cores += 1
    n_cores = get_cores()

    #print(n_cores)

    # assignment = partition(connex,n_cores)

    assert len(connections.keys()) - 1 in connections.keys()

    return axons, connections, inputs, outputs, n_cores




def partition(connectome, n_cores):
    """Creates adjacency list

    Uses the partitioning algorithm to partition the neurons in the network and return core assignments

    Parameters
    ----------
    network : dict
        Dictionary specifying neurons in the network. Key: Neuron Number Value: Synapse Weights
    ncores : int
        The number of cores peresent in the CRI system

    Returns
    -------
    dict
        Dictionary specifying neurons mapped to each core. Key: core number Value: tuple of (neuron number, core number)
    """
    networkConnectivity = connectome.get_part_format()

    # Create an adjacency list
    # Use the partitioning algorithm to partition the network and return the cluster assignments
    # it might be necessary to merge the axons and neurons into one network to get a good partitioning

    #TODO: get rid of this, it's just for some quick and dirty debugging
    #n_cores =2
    # No partitioning
    if n_cores == 1:
        membership = {k: 0 for k in range(len(networkConnectivity))}
        #return {k: 0 for k, v in connectome.connectomeDict.items()}
         #return {k: 0 for k, v in axons.items()},{k: 0 for k, v in connections.items()}

    # Partition the network
    else:
        if (partLoad):
            membership = netPart(data = networkConnectivity,n_clusters = n_cores) #n_clusters is just the number of cores to partition the network across
        else:
            logging.error("partitioning library failed to load, multicore partitioning skipped")

    connectome.apply_partition(membership)
    #breakpoint()


def external_input_optimization():
    return
    # TODO

"""
def map_to_hbm(axons, network, input, assignment, n_cores):

    # Should operate indpendent of the number of clusters
    # Return a dict of hbm ptrs and data for each cluster

    hbm = {}

    length = 0
    for i in axons:
        length += len(axons[i])
    for i in network:
        length += len(network[i])

    print("LENGTH ", length)

    for i in range(n_cores):
        cluster_nodes = {k: v for k, v in assignment.items() if v == i}
        # print('here')
        # print(cluster_nodes)

        axon_ptrs = len(axons.keys()) * [None]
        neuron_ptrs = len(network.keys()) * [None]

        # ptrs = (len(network.keys()) + len(axons.keys())) * [None]
        counter = 0
        data = []
        print(len(axons))
        # Create the external pointers section of the HBM
        for neuron in range(len(axons)):
            next = counter + int(np.ceil(len(axons[neuron]) / DATA_PER_ROW)) - 1
            if next < counter:
                next = counter
            axon_ptrs[neuron] = (counter, next)
            counter = counter + int(np.ceil(len(axons[neuron]) / DATA_PER_ROW))

        # If there we need more pointers to fit to the data_per_row size, append
        # pointers pointing to the empty data row

        if len(axons) % DATA_PER_ROW != 0:
            for j in range(DATA_PER_ROW - len(axons) % DATA_PER_ROW):
                axon_ptrs.append((counter, counter))
            counter += 1

        # print(axon_ptrs)

        # Create the internal pointers section of the hbm
        for neuron in range(len(network)):

            if assignment[neuron] == i:
                next = counter + int(np.ceil(len(network[neuron]) / DATA_PER_ROW)) - 1
                if next < counter:
                    next = counter
                neuron_ptrs[neuron] = (counter, next)
                counter = counter + int(np.ceil(len(network[neuron]) / DATA_PER_ROW))

        # If there we need more pointers to fit to the data_per_row size, append
        # pointers pointing to the empty data row

        # print(len(neuron_ptrs), len(network))

        if len(network) % DATA_PER_ROW != 0:
            for j in range(DATA_PER_ROW - len(network) % DATA_PER_ROW):
                neuron_ptrs.append((counter, counter))
            counter += 1

        # print(len(axon_ptrs), len(neuron_ptrs))
        dt = np.dtype("object,object")

        ptrs = np.array(axon_ptrs + neuron_ptrs, dtype=dt).reshape((-1, DATA_PER_ROW))

        # Create the data section of the hbm For axons
        for neuron in range(len(axons)):

            row_data = []

            if len(axons[neuron]) <= DATA_PER_ROW:
                for k in range(len(axons[neuron])):
                    row_data.append((0, axons[neuron][k][0], axons[neuron][k][1]))

            else:
                for k in range(len(axons[neuron])):
                    row_data.append((0, axons[neuron][k][0], axons[neuron][k][1]))
                    if k != 0 and k % DATA_PER_ROW == 0:
                        data.append(row_data)
                        row_data = []

            if len(row_data) < DATA_PER_ROW:
                for k in range(len(row_data), DATA_PER_ROW):
                    row_data.append((0, 0, 0))

            data.append(row_data)

        data.append([(0, 0, 0) for j in range(DATA_PER_ROW)])

        # Create data section of HBM for neurons
        for neuron in range(len(network)):

            row_data = []

            for k in range(len(network[neuron])):
                row_data.append((0, network[neuron][k][0], network[neuron][k][1]))
                if len(row_data) % DATA_PER_ROW == 0:
                    data.append(row_data)
                    row_data = []
            if 0 < len(row_data) < DATA_PER_ROW:
                for k in range(len(row_data), DATA_PER_ROW):
                    row_data.append((0, 0, 0))

            if row_data:
                data.append(row_data)

        # Add an emtpy data row to the end for any neurons that aren't important
        data.append([(0, 0, 0) for j in range(DATA_PER_ROW)])

        hbm[i] = (ptrs, data)

    print("Axon Pointers:")
    for i in axon_ptrs:
        print(i)

    print("Neuron Pointers:")
    for i in neuron_ptrs:
        print(i)

    print("HBM DATA")
    for i in data:
        print(i)

    return hbm
"""

def map_to_hbm_fpga(connectome, n_cores, to_fpga=True):
    """Creates HBM Data Structure

    Creates a representation of the axon pointers, neuron pointers, and synapse weights in HBM memory

    Parameters
    ----------
    axons : dict
        Dictionary specifying axons in the network. Key: Axon Number Value: Synapse Weights
    network : dict
        Dictionary specifying neurons in the network. Key: Neuron Number Value: Synapse Weights
    assignment : dict
        Dictionary specifying neurons mapped to each core. Key: core number Value: tuple of (neuron number, core number)
    n_cores : int
        The number of cores peresent in the CRI system
    to_fpga : bool, optional
        This parameter is depracated and has no effect. (the default is True)

    Returns
    -------
    hbm : dict
        Dictionary specifying the structure of data in memory for each core. Key: core number Value: tuple of (pointer,data) where pointer is a numpy array of tuples representing offsets into hbm memory and data is a list of lists tuples representing synapses.
    """

    # Should operate indpendent of the number of clusters
    # Return a dict of hbm ptrs and data for each cluster

    hbm = {}

    rows_per_ptr = ceil(N_NG / DATA_PER_ROW)

    for i in range(n_cores):
        #cluster_nodes = {k: v for k, v in assignment.items() if v == i}

        current = 0

        axon_ptrs = []
        axon_data = []

        ############################
        # Handle the Axons
        ############################
        axons = connectome.get_axons()
        for axonKey in axons:
            axon = axons[axonKey]
            weights = axon.get_synapses()
            used_weights = np.ones(len(weights))
            num_rows = 0
             


            if not weights:
                logging.warning("Axon with no synapses detected")
                if sixteen_flag:
                    # if there is no weights we append NG_Num number of rows of empty synapses and a pointer to those rows
                    for i in range(rows_per_ptr):
                        axon_data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
                    axon_ptrs.append((current, current + (rows_per_ptr - 1)))
                    current += rows_per_ptr
                else:
                    axon_data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
                    axon_ptrs.append((current, current))
                    current += 1
                continue

            while sum(used_weights) != 0:
                row = np.zeros(N_NG).astype(object)
                for w in range(len(weights)):
                    if used_weights[w] != 0:
                        column = weights[w].get_postsynapticNeuron().get_coreTypeIdx() % N_NG
                        if row[column] == 0:
                            row[column] = (
                                0,
                                np.floor(weights[w].get_postsynapticNeuron().get_coreTypeIdx() / N_NG),
                                weights[w].get_weight(),
                            )
                            used_weights[w] = 0
                row = [r if r != 0 else (0, 0, 0) for r in row]
                if sixteen_flag:
                    num_rows += rows_per_ptr
                    temp = np.empty(len(row), dtype=object)
                    temp[:] = row
                    axon_row_list = temp.reshape(-1, DATA_PER_ROW)
                    axon_row_list = np.flip(axon_row_list, axis=0)
                    for axon_row in axon_row_list:
                        axon_data.append(axon_row.tolist())
                else:
                    num_rows += 1  # For the pointers
                    axon_data.append(row)
            if sixteen_flag:
                axon_ptrs.append((current, current + num_rows - 1))
                current += num_rows
            else:
                axon_ptrs.append((current, current + num_rows - 1))
                current += num_rows

        if sixteen_flag:
            if len(axon_ptrs) % N_NG != 0:
                # make sure the row is filled with axon pointers TODO: do we just care the row is full or should the number
                # of axon pointers be divisable by 16
                # The correct answer is the total number of axon pointers should be evenly divisable by N_NG
                for j in range(N_NG - len(axon_ptrs) % N_NG):
                    axon_ptrs.append((current, current + rows_per_ptr - 1))
                    current += rows_per_ptr
                    for i in range(rows_per_ptr):
                        axon_data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
        else:
            if len(axon_ptrs) % DATA_PER_ROW != 0:
                for j in range(DATA_PER_ROW - len(axon_ptrs) % DATA_PER_ROW):
                    axon_ptrs.append((current, current))
                    current += 1
                    axon_data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])

        dt = np.dtype("object,object")
        if sixteen_flag:
            axon_ptrs = np.array(axon_ptrs, dtype=dt).reshape((-1, DATA_PER_ROW))
        else:
            axon_ptrs = np.array(axon_ptrs, dtype=dt).reshape((-1, DATA_PER_ROW))

        #################################
        # Handle the Neurons
        #################################
        ptrs = []
        data = []
        neurons = connectome.get_neurons()
        for neuronKey in neurons:
            neuron = neurons[neuronKey]
            weights = neuron.get_synapses()
            used_weights = np.ones(len(weights))
            num_rows = 0

            if not weights:
                if sixteen_flag:
                    # if no weights create a pointer to NG_num rows of zero synapses
                    for i in range(rows_per_ptr):
                        data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
                    # data.append([(0,0,0) for r in range(DATA_PER_ROW)])
                    ptrs.append((current, current + (rows_per_ptr - 1)))
                    current += rows_per_ptr
                else:
                    data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
                    ptrs.append((current, current))
                    current += 1
                continue

            space_for_spike = False
            while sum(used_weights) != 0:
                # Create a row
                row = np.zeros(N_NG).astype(object)

                for w in range(len(weights)):
                    if used_weights[w] != 0:
                        column = weights[w].get_postsynapticNeuron().get_coreTypeIdx() % N_NG
                        if row[column] == 0:
                            row[column] = (
                                0,
                                np.floor(weights[w].get_postsynapticNeuron().get_coreTypeIdx() / N_NG),
                                weights[w].get_weight(),
                            )
                            used_weights[w] = 0

                row = [r if r != 0 else (0, 0, 0) for r in row]
                if sixteen_flag:
                    num_rows += rows_per_ptr
                else:
                    num_rows += 1  # For the pointers
                if (0, 0, 0) in row:
                    space_for_spike = True

                if not row:
                    logging.info("empty neuron row found")
                if sixteen_flag:
                    # num_rows += rows_per_ptr
                    temp = np.empty(
                        len(row), dtype=object
                    )  # numpy get's weird about reshaping object arrays
                    temp[:] = row
                    # print(temp)
                    neuron_row_list = temp.reshape(-1, DATA_PER_ROW)
                    neuron_row_list = np.flip(neuron_row_list, axis=0)
                    # print(axon_row_list)
                    # print(DATA_PER_ROW)
                    for neuron_row in neuron_row_list:
                        # print(axon_row.tolist())
                        data.append(neuron_row.tolist())
                else:
                    data.append(row)

            # No place for spike data, we add an additional empty row
            # TODO: is it sufficient to simply add one extra row or does in
            # need to be rows_per_ptr number of empty rows?
            # I'm pretty sure it needs to be rows_per_ptr number of empty rows
            if not space_for_spike:
                if sixteen_flag:
                    for i in range(rows_per_ptr):
                        data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
                    # data.append([(0,0,0) for r in range(DATA_PER_ROW)])
                    num_rows += rows_per_ptr
                else:
                    data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
                    num_rows += 1

            ptrs.append((current, current + num_rows - 1))
            current += num_rows

        if len(ptrs) % N_NG != 0:
            #print("need to pad neuron pointers")
            #print(ptrs)
            if sixteen_flag:
                for j in range(N_NG - len(ptrs) % N_NG):
                    ptrs.append((current, current + (rows_per_ptr - 1)))
                    current += rows_per_ptr
                    for i in range(rows_per_ptr):
                        data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])
            else:
                for j in range(DATA_PER_ROW - len(ptrs) % DATA_PER_ROW):
                    ptrs.append((current, current))
                    current += 1
                    data.append([(0, 0, 0) for r in range(DATA_PER_ROW)])

        dt = np.dtype("object,object")
        if sixteen_flag:
            ptrs = np.array(ptrs, dtype=dt).reshape((-1, DATA_PER_ROW))
        else:
            ptrs = np.array(ptrs, dtype=dt).reshape((-1, DATA_PER_ROW))

        data = axon_data + data


        hbm[i] = (axon_ptrs, ptrs, data)

        ptrs = np.vstack((axon_ptrs, ptrs))
        

       

    return hbm  # TODO this will need to be fixed for when we move to multicore, it currently only returns hbm for a single core


def compileNetwork(
    loadFile=False, connectome=None, inputs=None, outputs=None):
    """Creates simulation and FPGA data structures

    Creates a representation of the axon pointers, neuron pointers, and synapse weights in HBM memory both in the format used to produce the commands to program the actual FPGA and in the format expected by the hardware simulator.

    Parameters
    ----------
    loadFile : bool
        if True network topology is read from files on disk
    connectome : obj
        object encapsulating axons and connections
    axons : dict
        Dictionary specifying axons in the network. Key: axon number Value: Synapse Weights
    connections : dict
        Dictionary specifying neurons in the network. Key: Neuron Number Value: Synapse Weights
    inputs : dict
        Dictionary specifying inputs to the network. Key, Time Step Value, axon
    outputs : dict
        TODO: I'm not sure what the outputs are for. I belive it's unused
    Returns
    -------
   input : dict
        Dictionary specifying inputs to the network. Key, Time Step Value, TODO are these axons or neurons
    hbm : dict
        Dictionary specifying the hbm structure for each core expected by the hardware simulator. Key: core number Value: tuple of (pointer,data) where pointer is a numpy array and data is a list of lists of tuples.
    outputs : dict
        TODO: I'm not sure what the outputs are for
    axonLength : int
        number of axons specified in the network

"""
    if loadFile:
        axons, network, input, output, n_cores = load_network()
    else:
        if connectome == None or outputs == None:
            raise ValueError(
                "compileNetwork was called with loadFile set to False and one or more input dicitonary was empty"
            )
        # normally load_network returns n_cores, instead we'll just do it explicetly

        n_cores = get_cores()

    partition(connectome, n_cores)
    #generate_local_neurons(axons,connections,axonAssignments , neuornAssignments, n_cores)
    #you're going to have to generate an hbm mapping for every core
    fpga = map_to_hbm_fpga(connectome, n_cores)
    axonLength = len(connectome.get_axons())
    if loadFile:
        return outputs, fpga, axonLength
    else:
        return fpga, axonLength


def main():
    compileNetwork()


if __name__ == "__main__":
    main()



from cri_simulations.compile_network import compileNetwork
from cri_simulations.FPGA_Execution.fpga_compiler import fpga_compiler
import subprocess
import numpy as np
import time
import functools
import operator
import logging
from cri_simulations.config import *
from cri_simulations.FPGA_Execution.fpga_controller import write_parameters, clear, step_input, read, execute, write_synapse_row, input_user, flush_spikes, clear_read_buffer
class network:
    """A class for creating networks to be run on the CRI hardware

    Attributes
    ----------
    inputs : dict
        Dictionary specifying inputs to the network. Key, Time Step Value, TODO are these axons or neurons
    hbmRecord : dict
        Dictionary specifying the hbm structure for each core. Key: core number Value: tuple of (pointer,data) where pointer is a numpy array and data is a list of lists of tuples.
    outputs : dict
        TODO: I'm not sure what the outputs are for
    axonLength : int
        number of axons specified in the network
    self.numNeuron : int
        number of neurons specified in the network
    """

    def __init__(self, connectome, outputs, config, simDump = False, coreOveride = 0):
        self.coreOveride = coreOveride #TODO:this is for debugging
        self.hbm, self.numAxon = compileNetwork(
            loadFile=False,
            connectome = connectome,
            outputs=outputs
            
        )
        # TODO: this will need to be generalized for multiple networks
        #print(self.hbm.keys())
        axon_ptrs, ptrs, data = self.hbm[1] #This starts at 1 for whatever reason TODO: fix this so it's zero indexed
        self.num_inputs = len(connectome.get_axons()) #number of axons in the network
        self.num_outputs = len(connectome.get_neurons())
        self.outputs = outputs
        self.numNeurons=len(connectome.get_neurons())


        #print('connection points')
        #print(len(connections))
        self.compiledNetwork = fpga_compiler(
            (axon_ptrs, ptrs, data), self.numNeurons, self.outputs, coreID = self.coreOveride
        )
        self.stepNum = None
        #self.maxTimeStep = max(inputs.keys()) #TODO: Change this
        self.neuronType = config['neuron_type']
        self.voltageThresh = config['global_neuron_params']['v_thr']
        self.simDump = simDump
        if simDump:
            self.cmdDump = []

    def initalize_network(self):
        if self.simDump:
            axon_ptrs, neuron_ptrs, synapses = self.compiledNetwork.create_script("test_config", simDump = True)
            self.cmdDump.append(axon_ptrs)
            self.cmdDump.append(neuron_ptrs)
            self.cmdDump.append(synapses)
            #write_parameters(3, self.voltageThresh, n_outputs=length(self.connections), n_inputs=length(self.axons))
            param_cmd = write_parameters(3, self.voltageThresh, n_outputs=self.num_outputs, n_inputs=self.num_inputs, simDump = True, coreID = self.coreOveride)
            self.cmdDump.append(param_cmd)
            clear_cmd = clear(self.numNeurons, simDump = True, coreID = self.coreOveride)
            self.cmdDump.append(clear_cmd)
            self.stepNum = 0
        else:
            clear_read_buffer(coreID = self.coreOveride) #make sure no old data is still in the read buffer
            self.compiledNetwork.create_script("test_config")
            subprocess.call(["cp", "test_config_script.txt", "test_config_script.sh"])

            subprocess.call(["bash", "test_config_script.sh"])
            # subprocess.call(['sudo', 'bash', 'test_input_script.sh'])

      
            # write_parameters(neron_model, threshold, n_outputs, n_inputs)
            #TODO These parameters need to be class attributes
            write_parameters(3, self.voltageThresh, n_outputs=self.num_outputs, n_inputs=self.num_inputs) #TODO: generalize past I&F
            #write_parameters(
            #    3, self.voltageThresh, n_outputs=40, n_inputs=5
            #)
            clear(self.numNeurons)
            self.stepNum = 0

    def run_step(self,inputs,  membranePotential = False):
        if True: #self.stepNum <= self.maxTimeStep:
            if self.simDump:
                input_cmd = input_user(inputs, numAxons = self.num_inputs,  simDump = True,coreID = self.coreOveride)
                self.cmdDump.append(input_cmd)
                #time.sleep(4)
                execute_cmd = execute(simDump = True,coreID = self.coreOveride)
                self.cmdDump.append(execute_cmd)
                #time.sleep(8)
                read_cmd = read(self.numNeurons, simDump = True,coreID = self.coreOveride)
                self.cmdDump.append(read_cmd)
                self.stepNum = self.stepNum +1
                #return result
            else:
                input_user(inputs, numAxons = self.num_inputs, coreID = self.coreOveride)
                #time.sleep(4)
                execute(coreID = self.coreOveride)
                #time.sleep(8)
                spikes = flush_spikes(coreID = self.coreOveride)
                self.stepNum = self.stepNum +1
                if membranePotential:
                    result = read(self.numNeurons,coreID = self.coreOveride)  # TODO make read return the values instead of just printing to terminal
                    return result, spikes
                else: 
                    return spikes
        else:
            logging.warning("run_step did nothing, network has already finished all timesteps")

    def read_synapse(self,preIndex, postIndex, axonFlag = False):
        if self.simDump:
            logging.error("read_synapse commands not added to simDump")
        else:
            #TODO: it optionally might be nice to be able to read from the actual hardware
            if axonFlag:
                pntrs = self.hbm[1][0]
            else:
                pntrs = self.hbm[1][1]
    
            synapseRange = pntrs.flatten()[preIndex]
            synapses = self.hbm[1][2][synapseRange[0]:synapseRange[1]+1]
            columnIdx = postIndex % DATA_PER_ROW #This will correctly find you the column
            rowIdx = np.floor(postIndex / N_NG) #This will tell you what the "within neuron group


            search_synapses = [idx for idx,i in enumerate(synapses) if i[columnIdx][0] == 0 and i[columnIdx][1] == rowIdx and i[columnIdx] != (0,0,0) ]
            if (len(search_synapses) != 1):
                raise ValueError('0 or multiple valid synapses found')
            synapseIdx = (search_synapses[0], columnIdx)
            return synapses[synapseIdx[0]][synapseIdx[1]]

    def sim_flush(self, file):
        dmpStr = ""
        if self.simDump:
            cmdDump = functools.reduce(operator.concat, self.cmdDump)
            with open(file, 'w') as f:
                for item in cmdDump:
                    f.write("%s\n" % item)
                    dmpStr = dmpStr + item + "\n"
            return dmpStr
        else:
            raise Exception("simdump was not set to True at object creation. No commands to flush")

    def write_synapse(self,preIndex, postIndex, weight, axonFlag = False):
#TODO: combine read and write synapse to avoid duplicating code
        if self.simDump:
            logging.warning("write_synapse commands not added to simDump, not implemented yet")
        else:
            if axonFlag:
                pntrs = self.hbm[1][0]
            else:
                pntrs = self.hbm[1][1]

            synapseRange = pntrs.flatten()[preIndex]
            synapses = self.hbm[1][2][synapseRange[0]:synapseRange[1]+1]
            columnIdx = postIndex % DATA_PER_ROW #This will correctly find you the column
            rowIdx = np.floor(postIndex / N_NG) #This will tell you what the "within neuron group index is"


            search_synapses = [idx for idx,i in enumerate(synapses) if i[columnIdx][0] == 0 and i[columnIdx][1] == rowIdx and i[columnIdx] != (0,0,0) ]
            if (len(search_synapses) != 1):
                raise ValueError('0 or multiple valid synapses found')
            synapseIdx = (search_synapses[0], columnIdx)
            oldSynapse = synapses[synapseIdx[0]][synapseIdx[1]]
            row = synapses[synapseIdx[0]]
            row[columnIdx] = (oldSynapse[0],oldSynapse[1],weight)
            write_synapse_row(synapseRange[0]+synapseIdx[0], row, simDump = False)
           #This appears to actually update the values in hbm
                



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cri_simulations', 'cri_simulations.FPGA_Execution']

package_data = \
{'': ['*']}

install_requires = \
['PyMetis>=2020.1,<2021.0',
 'bidict>=0.22.0,<0.23.0',
 'click>=8.1.3,<9.0.0',
 'confuse>=1.7.0,<2.0.0',
 'fbpca>=1.0,<2.0',
 'k-means-constrained>=0.7.1,<0.8.0',
 'matplotlib>=3.5.2,<4.0.0',
 'metis>=0.2a5,<0.3',
 'networkx>=2.8.3,<3.0.0',
 'numpy>=1.18',
 'scipy>=1.8.1,<2.0.0',
 'sklearn>=0.0,<0.1']

setup_kwargs = {
    'name': 'cri-simulations',
    'version': '0.1.2',
    'description': 'Software for interacting with the CRI neuromorphic hardware',
    'long_description': None,
    'author': 'Justin Frank',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
